```@meta
CurrentModule = Oscar
```

```@setup oscar
using Oscar
```

```@contents
Pages = ["ca_binomial_ideals.md"]
```

# Binomial Primary Decomposition

## Introduction

A binomial is a polynomial consisting of at most two terms. A binomial ideal
is an ideal which can be generated by binomials. A binomial primary decomposition
is a primary decomposition of a binomial ideal into primary ideals which are binomial
as well. In this section, we discuss functionality for computing such decompositions.

The crucial first step of the binomial primary decomposition algorithm is to decompose
a given binomial ideal into cellular binomial ideals. Here, recall
that a proper binomial ideal $I$ of a polynomial ring $R = K[x_1, \dots, x_n]$
is called cellular if each variable $x_i$ is either a nonzerodivisor or nilpotent modulo $I$.

!!! note
    Current status: characteristic zero, pure binomial ideals

Papers offering details on theory and algorithms as well as examples include:

- [ES96](@cite)
- [OMdCS00](@cite)
- [Kah10](@cite)
- [EM16](@cite)
- [EM19](@cite)

## Basic Tests

### Binomiality Test

```@docs
isbinomial(f::MPolyElem)
isbinomial(I::MPolyIdeal)
```
```@repl oscar
R, (x, y, z) = PolynomialRing(QQ, ["x", "y", "z"])
f = 2*x+y
isbinomial(f)
J = ideal(R, [x^2-y^3, z^2])
isbinomial(J)
```

### Cellularity Test

```@docs
iscellular(I::MPolyIdeal)
```

### Unitality Test

```@docs
isunital(I::MPolyIdeal)
```



## Cellular Decomposition

```@docs
cellular_decomposition(I::MPolyIdeal)
```

## Primary Decomposition of Cellular Ideals

```@docs
cellular_hull(I::MPolyIdeal)
```

```@docs
cellular_associated_primes
```

```@docs
cellular_minimal_associated_primes
```

```@docs
cellular_primary_decomposition(I::MPolyIdeal)
```

## Primary Decomposition of Binomial  Ideals

```@docs
binomial_primary_decomposition(I::MPolyIdeal)
```


