#TODO: include more examples with nontrivial lineality space

@testset "Polyhedron" begin

    pts = [1 0 0; 0 0 1]'
    Q0 = convex_hull(pts)
    Q1 = convex_hull(pts, [1 1])
    Q2 = convex_hull(pts, [1 1], [1 1])
    square = cube(2)
    C1 = cube(2, 0, 1)
    Pos=Polyhedron([-1 0 0; 0 -1 0; 0 0 -1],[0,0,0])
    point = convex_hull([0 1 0])
    s = simplex(2)

    @testset "core functionality" begin
        @test nvertices(Q0) == 3
        @test nvertices.(faces(Q0,1)) == [2,2,2]
        @test length.(lattice_points(Q0)) == [2;2;2]
        @test isfeasible(Q0)
        @test issmooth(Q0)
        @test isnormal(Q0)
        @test isbounded(Q0)
        @test isfulldimensional(Q0)
        @test f_vector(Q0) == [3,3]
        @test intersect(Q0, Q0) == Q0
        @test Q0+Q0 == minkowski_sum(Q0, Q0)
        @test f_vector(Pos) == [1,3,3]
        @test codim(square) == 0
        @test codim(point) == 3
        @test !isfulldimensional(point)
        @test nrays(recession_cone(Pos)) == 3
        @test vertices_as_point_matrix(2*point) == [0 2 0]
        @test vertices_as_point_matrix([0,1,0] + point) == [0 2 0]
        @test length(collect(rays(Pos))) == 3
    end

    @testset "linear programs" begin
        LP1 = LinearProgram(square,[1,3])
        LP2 = LinearProgram(square,[2,2]; k=3, convention = :min)
        LP3 = LinearProgram(Pos,[1,2,3])

        @test solve_lp(LP1)==(4,[1,1])
        @test solve_lp(LP2)==(-1,[-1,-1])
        @test string(solve_lp(LP3))=="(inf, nothing)"
    end

    @testset "volume" begin
        @test volume(square) == 4
        @test normalized_volume(square) == 8
        @test normalized_volume(s) == 1
    end

    @testset "standard_constructions" begin
        p = upper_bound_theorem(4,8)
        @test p.pm_polytope.F_VECTOR == [8, 28, 40, 20]
        A = archimedean_solid("cuboctahedron")
        @test count(F -> nvertices(F) == 3, faces(A, 2)) == 8
        # due to GLIBCXX issues with the outdated julia-shipped libstdc++
        # we run this only where recent CompilerSupportLibraries are available
        if VERSION >= v"1.6"
            C = catalan_solid("triakis_tetrahedron")
            @test count(F -> nvertices(F) == 3, faces(C, 2)) == 12
        end
        nc = normal_cone(square, 1)
        @test rays_as_point_matrix(nc) == [1 0; 0 1]
    end

end
